/*
 * Copyright 2009 Nicolas LE GRIEL.
 * 
 * Les sources qui constituent ce projet nlegriel-dvp-defi5 de mme que 
 * la documentation associe sont la proprit de leur auteur.
 * 
 * Je donne mon accord au site developpez.com pour l'utilisation de 
 * tout ou partie des sources et de la documentation de ce projet dans 
 * les pages developpez.com
 */
package com.developpez.nlegriel.defi5;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Matrice carre d'ordre 3  coefficients entiers utilise pour les
 * oprations de rotations sur des cubes.
 *
 * @author Nicolas LE GRIEL
 */
public class IntegerMatrix3x3 {
	/** Identit. */
	public final static IntegerMatrix3x3 ID = new IntegerMatrix3x3(
			new int[][] {
					new int[] {1, 0, 0},
					new int[] {0, 1, 0},
					new int[] {0, 0, 1},
			}
	);
	
	/** Rotation d'angle pi/2 autour de l'axe des abscisses. */
	public final static IntegerMatrix3x3 X = new IntegerMatrix3x3(
			new int[][] {
					new int[] {1, 0, 0},
					new int[] {0, 0, -1},
					new int[] {0, 1, 0},
			}
	);
	
	/** Rotation d'angle pi/2 autour de l'axe des ordonnes. */
	public final static IntegerMatrix3x3 Y = new IntegerMatrix3x3(
			new int[][] {
					new int[] {0, 0, 1},
					new int[] {0, 1, 0},
					new int[] {-1, 0, 0},
			}
	);
	
	/** Rotation d'angle pi/2 autour de l'axe des cotes. */
	public final static IntegerMatrix3x3 Z = new IntegerMatrix3x3(
			new int[][] {
					new int[] {0, -1, 0},
					new int[] {1, 0, 0},
					new int[] {0, 0, 1},
			}
	);
	
	public final static IntegerMatrix3x3 ZZ = Z.multiply(Z);
	public final static IntegerMatrix3x3 ZZZ = ZZ.multiply(Z);
	public final static IntegerMatrix3x3 YYY = Y.multiply(Y.multiply(Y));
	public final static IntegerMatrix3x3 ZX = Z.multiply(X);
	public final static IntegerMatrix3x3 ZZX = ZZ.multiply(X);
	public final static IntegerMatrix3x3 ZZZX = ZZZ.multiply(X);
	public final static IntegerMatrix3x3 YX = Y.multiply(X);
	public final static IntegerMatrix3x3 YYYX = YYY.multiply(X);
	public final static IntegerMatrix3x3 XX = X.multiply(X);
	public final static IntegerMatrix3x3 ZXX = ZX.multiply(X);
	public final static IntegerMatrix3x3 ZZXX = ZZX.multiply(X);
	public final static IntegerMatrix3x3 ZZZXX = ZZZX.multiply(X);
	public final static IntegerMatrix3x3 YXX = YX.multiply(X);
	public final static IntegerMatrix3x3 YYYXX = YYYX.multiply(X);
	public final static IntegerMatrix3x3 XXX = XX.multiply(X);
	public final static IntegerMatrix3x3 ZXXX = ZXX.multiply(X);
	public final static IntegerMatrix3x3 ZZXXX = ZZXX.multiply(X);
	public final static IntegerMatrix3x3 ZZZXXX = ZZZXX.multiply(X);
	public final static IntegerMatrix3x3 YXXX = YXX.multiply(X);
	public final static IntegerMatrix3x3 YYYXXX = YYYXX.multiply(X);
	
	/** Ensemble des isomtries positives laissant le cube de centre (0, 0, 0) invariant. */
	public final static List<IntegerMatrix3x3> ALL_ROTATIONS;
	static {
		List<IntegerMatrix3x3> rotations = new ArrayList<IntegerMatrix3x3>();
		rotations.add(ID); rotations.add(Z); rotations.add(ZZ); rotations.add(ZZZ); rotations.add(Y); rotations.add(YYY); 
		rotations.add(X); rotations.add(ZX); rotations.add(ZZX); rotations.add(ZZZX); rotations.add(YX); rotations.add(YYYX); 
		rotations.add(XX); rotations.add(ZXX); rotations.add(ZZXX); rotations.add(ZZZXX); rotations.add(YXX); rotations.add(YYYXX); 
		rotations.add(XXX); rotations.add(ZXXX); rotations.add(ZZXXX); rotations.add(ZZZXXX); rotations.add(YXXX); rotations.add(YYYXXX);
		ALL_ROTATIONS = Collections.unmodifiableList(rotations);
	}

	/** Reprsentation interne des coefficients de la matrice. */
	private int[][] coefficients = new int[3][3];
	
	
	/** Fournit une matrice nulle (tous les coefficients gaux  0). */
	public IntegerMatrix3x3() {
	}
	
	/**
	 * Initialise la matrice en recopiant les coefficients fournis. 
	 * 
	 * @param coefficients tableau de dimension 3x3
	 * 
	 * @throws IllegalArgumentException si les dimensions ne
	 * sont pas respectes ou s'il y a des rfrences <code>null</code>
	 * dans le tableau fourni en argument 
	 */
	public IntegerMatrix3x3(int[][] coefficients) {
		if (coefficients == null)
			throw new IllegalArgumentException("Rfrence nulle fournie en argument");
		if (coefficients.length != 3)
			throw new IllegalArgumentException("Nombre de lignes diffrent de 3");
		if (coefficients[0] == null || coefficients[1] == null || coefficients[2] == null)
			throw new IllegalArgumentException("Ligne 'null'");
		if (coefficients[0].length != 3 || coefficients[1].length != 3 || coefficients[2].length != 3)
			throw new IllegalArgumentException("Nombre de colonnes diffrent de 3");

		for (int i=0; i<3; i++)
			System.arraycopy(coefficients[i], 0, this.coefficients[i], 0, 3);
	}
	
	/**
	 * Calcule la nouvelle coordone d'un cube du puzzle par application de
	 * la transformation linaire dfinie par la prsente matrice.
	 * 
	 * @param cube contient trois coordonnes qui forment un vecteur
	 * 
	 * @return un nouveau cube dont les coordonnes sont calcules  
	 * partir de la prsente matrice
	 */
	public Cube operate(Cube cube) {
		Cube result = new Cube();
		
		result.x = coefficients[0][0]*cube.x + coefficients[0][1]*cube.y + coefficients[0][2]*cube.z;
		result.y = coefficients[1][0]*cube.x + coefficients[1][1]*cube.y + coefficients[1][2]*cube.z;
		result.z = coefficients[2][0]*cube.x + coefficients[2][1]*cube.y + coefficients[2][2]*cube.z;
		
		return result;
	}
	
	/**
	 * Calcule le projet de la matrice courante par la matrice fournie en
	 * argument.
	 * 
	 * @param m membre de droite du produit
	 * 
	 * @return une nouvelle matrice contenant le rsultat du produit matriciel
	 */
	public IntegerMatrix3x3 multiply(IntegerMatrix3x3 m) {
		int[][] result = new int[3][3];
		
		for (int i=0; i<3; i++)
			for (int j=0; j<3; j++)
				for (int k=0; k<3; k++)
					result[i][j] += coefficients[i][k] * m.coefficients[k][j];
		
		return new IntegerMatrix3x3(result);
	}
	
	/**
	 * Teste l'galit des matrices coefficient par coefficient.
	 * 
	 * {@inheritDoc}
	 */
	@Override
	public boolean equals(Object o) {
		if ( o == null || ! (o instanceof IntegerMatrix3x3) )
			return false;
		
		IntegerMatrix3x3 m = (IntegerMatrix3x3) o;
		
		for (int i=0; i<3; i++)
			for (int j=0; j<3; j++)
				if (coefficients[i][j] != m.coefficients[i][j])
					return false;
				
		return true;
	}
}
