/*
 * Copyright 2009 Nicolas LE GRIEL.
 * 
 * Les sources qui constituent ce projet nlegriel-dvp-defi5 de mme que 
 * la documentation associe sont la proprit de leur auteur.
 * 
 * Je donne mon accord au site developpez.com pour l'utilisation de 
 * tout ou partie des sources et de la documentation de ce projet dans 
 * les pages developpez.com
 */
package com.developpez.nlegriel.defi5;

import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringWriter;
import java.util.Random;
import java.util.Scanner;

import com.developpez.nlegriel.defi5.view.LauncherWindow;

/**
 * Lanceur de l'application.
 *
 * @author Nicolas LE GRIEL
 */
public class Launcher {
	private final static Random random = new Random(System.currentTimeMillis());
	
	public final static String DEFAULT_OUTPUT_EXTENSION = ".out.txt";
	
	/** Statut de l'algorithme de recherche de solution. */
	public static enum Status {
		STARTING, RUNNING, FAILURE, SUCCESS
	}
	
	private Status status = Status.STARTING;
	
	private Puzzle puzzle;
	
	private Reader input;
	private StringWriter stringWriter;
	private PrintWriter output;
	
	/** Utilis pour grer un chronomtre d'excution de l'algorithme de recherche. */
	private long t0;
	
	
	/**
	 * Excute l'application en mode <i>console</i> ou en mode <i>fentre</i> selon
	 * qu'elle ait t lance avec ou sans arguments sur la ligne de commande.
	 * 
	 * @param arguments peut tre vide ou contenir un ou deux chemins de fichiers
	 * (absolus ou relatifs) rcuprs par le constructeur {@link #Launcher(String...)}.
	 * 
	 * @throws IOException si un fichier de sortie n'a pu tre cr
	 */
	public static void main(String... arguments) throws IOException {
		Launcher launcher = new Launcher(arguments);
		if ( launcher.getReader() == null || "".equals(launcher.getReader()) )
			new LauncherWindow(launcher).setVisible(true);	
		else {
			launcher.setT0(System.currentTimeMillis());
			launcher.initPuzzle();
			launcher.execute();
			if (launcher.getStringWriter() != null)
				System.out.println(launcher.getStringWriter().toString());
		}
	}

	/**
	 * Fournit le flux cens provenir d'un fichier de dfinition de
	 * puzzle.
	 */
	public Reader getReader() {
		return input;
	}
	
	public StringWriter getStringWriter() {
		return stringWriter;
	}
	
	/**
	 * Fournit le flux de sortie devant tre produit par la mthode 
	 * {@link #execute()} et contenant le rsultat de l'analyse du 
	 * flux indiqu par le champ {@link #input}.
	 */
	public PrintWriter getWriter() {
		return output;
	}

	/**
	 * Fournit une description de la figure  tudier ainsi que diverses
	 * informations sur l'analyse en cours d'excution.
	 */
	public Puzzle getPuzzle() {
		return puzzle;
	}

	/**
	 * Fournit l'tat de l'algorithme de recherche de solution au cube de Soma pour
	 * la figure  tudier.
	 */
	public Status getStatus() {
		return status;
	}
	
	/**
	 * Rinitialise l'instant <i>t0</i> servant de base pour l'valuation du 
	 * temps d'xcution de l'algorithme de recherche.
	 *  
	 * @param t0 date courante du systme en millisecondes
	 */
	public void setT0(long t0) {
		this.t0 = t0;
	}

	/**
	 * Analyse les arguments transmis depuis la ligne de commande.
	 * <p>
	 * Trois cas de figure sont autoriss :
	 * <ul>
	 * 	<li>Aucun argument</li>
	 * 	<li>Un argument fournissant un chemin (absolu ou relatif) de fichier d'entre</i>
	 * 	<li>
	 * 		Deux arguments fournissant deux chemins (absolus ou relatifs) de fichiers. 
	 * 		Le premier pour l'entre et le second pour la sortie.
	 *	</i>
	 * </ul>
	 * Les arguments supplmentaires sont ignors.
	 * 
	 * @param arguments arguments de la ligne de commande
	 * 
	 * @throws IOException si un fichier de sortie n'a pu tre cr
	 */
	public Launcher(String... arguments) throws IOException {
		if (arguments == null)
			return;
		else if (arguments.length == 1) {
			input = new FileReader(arguments[0]);
			stringWriter = new StringWriter();
			output = new PrintWriter(stringWriter);
		}
		else if (arguments.length >= 2) {
			input = new FileReader(arguments[0]);
			output = new PrintWriter(new FileWriter(arguments[1]));
		}
	}
	
	/**
	 * Initialise le puzzle  partir de la dfinition de puzzle
	 * fournie en flux d'entre.
	 * 
	 * @throws RuntimeException si un problme a t dtect durant la
	 * lecture des donnes d'entre.
	 */
	public void initPuzzle() {
		try {
			puzzle = new Puzzle();
			puzzle.loadFromScanner(new Scanner(input));
			puzzle.resetUncoverdCubes();
			puzzle.resetUnusedParts();
		}
		catch (RuntimeException e) {
			output.println("Erreur survenue durant la lecture des donnes d'entres");
			e.printStackTrace(output);
			output.close();
			throw e;
		}
	}
	
	/**
	 * Excute l'algorithme d'analyse du puzzle indiqu par le 
	 * champ {@link #input}.
	 * 
	 * @throws IOException en cas d'impossibilit d'crire sur le flux de
	 * sortie
	 */
	public void execute() throws IOException {
		if (t0 == 0)
			t0 = System.currentTimeMillis();
		
		try {
			if (puzzle == null)
				initPuzzle();
		}
		catch (Exception e) {
			// La mthode initPuzzle a produit son propre message d'erreur et a ferm
			// le flux de sortie :
			return;
		}
			
		try {
			this.status = Status.FAILURE;
			if (searchConfiguration(puzzle))
				this.status = Status.SUCCESS;
			
			if (status == Status.SUCCESS)
				output.println("Une solution a t trouve pour la figure propose.");
			else
				output.println("La figure propose n'a pas de solution.");
			
			output.println("Dure du traitement de recherche : "+(System.currentTimeMillis() - t0)+" ms");
		}
		catch (Exception e) {
			output.println("Erreur survenue durant le traitement.");
			e.printStackTrace(output);
		}
		finally {
			output.close();
		}
	}

	/**
	 * Recherche <b>une</b> solution au puzzle par un principe de <i>back tracking</i>. 
	 * 
	 * @param puzzle rfrence vers la situation du puzzle. Les champs 
	 * {@link Puzzle#uncoveredCubes} et {@link Puzzle#unusedParts} sont modifis
	 * au cours de la recherche de solution.
	 * 
	 * @return <code>true</code> ds qu'une solution a t trouve au puzzle
	 * 
	 * @throws IOException en cas d'impossibilit d'crire sur le flux de sortie
	 */
	private boolean searchConfiguration(Puzzle puzzle) throws IOException {
		int randomPosition = random.nextInt(puzzle.uncoveredCubes.size());
		Cube uncoveredCube = puzzle.uncoveredCubes.get(randomPosition);
		
		int i = 0;
		while (i < puzzle.unusedParts.size()) {
			PuzzlePart part = puzzle.unusedParts.get(i);
			puzzle.unusedParts.remove(part);
			if (puzzle.unusedParts.size() == 6) {
				/*
				 *  Si la log est dans un fichier, on peut voir les lignes ci-dessous s'crire 
				 *  au fur et  mesure du traiement (si le traitement est long).
				 *  
				 *  Pour le mode IHM, il faudrait mettre en place un thread pour voir le 
				 *  message s'afficher en temps rel sur la fentre.
				 */
				output.printf("Solutions possibles tudies : %2.2f %%\n", 100*(double)i/(puzzle.unusedParts.size()+1));
				output.flush();
			}
			
			for (Cube[] zeroPartPositions : part.getPositions()) {
				// Faire passer la pice par le cube de la figure actuellement slectionn :
				Cube[] translatedPartPositions = new Cube[zeroPartPositions.length];
				for (int j=0; j<translatedPartPositions.length; j++)
					translatedPartPositions[j] = zeroPartPositions[j].translate(uncoveredCube);
				
				/*
				 *  Est-ce que les cubes occups par cette position de la pice courante 
				 *  sont tous des cubes non couverts ?
				 */
				boolean isPositionOk = true;
				for (Cube cube : translatedPartPositions)
					if ( ! puzzle.uncoveredCubes.contains(cube) ) {
						isPositionOk = false;
						break;
					}
				
				if (isPositionOk) {
					// On vient de russir  placer une pice du puzzle. On le note :
					for (Cube cube : translatedPartPositions) {
						puzzle.uncoveredCubes.remove(cube);
						puzzle.coveredCubes.add(new CubeCover(cube, part));
					}
					
					// Si on tait sur la dernire pice  placer, cela signifie qu'on a 
					// trouv une solution :
					if (puzzle.unusedParts.isEmpty())
						return true;
					
					// On peut maintenant essayer de placer les autres pices du puzzle :
					if (searchConfiguration(puzzle))
						// Si on arrive ici, c'est que la suite du traitement a permis 
						// de trouver une solution au puzzle :
						return true;
					else {
						/*
						 *  Si on arrive ici, c'est que le placement actuel de la pice ne 
						 *  permet pas de rsoudre le puzzle. On essaye donc de placer 
						 *  autrement la pice avant d'essayer une autre pice.
						 *  
						 *  Avant toute chose, il faut "librer" les cubes actuellement
						 *  couverts par la position courante de la pice courante.
						 */
						for (Cube cube : translatedPartPositions) {
							puzzle.uncoveredCubes.add(cube);
							puzzle.coveredCubes.remove(new CubeCover(cube, part));
						}
					}
				}
			}
			
			// La pice courante n'a pas permis de trouver une solution, on passe  la suivante :
			puzzle.unusedParts.add(i, part);
			i++;
		}
		
		// Si on arrive ici c'est parce que le contenu de la liste puzzle.unusedParts ne
		// permet pas de trouver une solution :
		return false;
	}
}
