/*
 * Copyright 2009 Nicolas LE GRIEL.
 * 
 * Les sources qui constituent ce projet nlegriel-dvp-defi5 de mme que 
 * la documentation associe sont la proprit de leur auteur.
 * 
 * Je donne mon accord au site developpez.com pour l'utilisation de 
 * tout ou partie des sources et de la documentation de ce projet dans 
 * les pages developpez.com
 */
package com.developpez.nlegriel.defi5;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Scanner;
import java.util.Set;


/**
 * Ensemble des informations permettant de reprsenter un puzzle du 
 * problme du cube de SOMA.
 * <p>
 * Un objet de cette classe contient aussi des informations modifiables
 * pouvant tre utilise au cours d'un algorithme de recherche de solution
 * de recouvrement de la figure par les pices du puzzle.
 *
 * @author Nicolas LE GRIEL
 */
public class Puzzle {
	/** Syntaxe des lignes d'un fichier de dfinition de figure. */
	private final static String SOURCE_LINE_PATTERN = "\\d,\\d,\\d";
	
	public final static int CUBES_NUMBER = 27;
	
	/** Cubes de la figure  analyser. */
	public Cube[] cubes = new Cube[CUBES_NUMBER];
	
	/**
	 * Reprsente l'ensemble des cube de la figure tudie qui n'ont pas
	 * encore t recouverts par une pice du puzzle.
	 */
	public List<Cube> uncoveredCubes = new ArrayList<Cube>();
	
	/**
	 * Pices du cube de Soma non encore utilises pour recouvrir la figure en
	 * cours d'analyse.
	 */
	public List<PuzzlePart> unusedParts = new ArrayList<PuzzlePart>();
	
	/**
	 * Utilise pour pouvoir visualiser graphiquement la position des
	 * pices dans le puzzle.
	 */
	public List<CubeCover> coveredCubes = new ArrayList<CubeCover>();
	
	
	/**
	 * Rinitialise le champ {@link #uncoveredCubes}. 
	 */
	public void resetUncoverdCubes() {
		uncoveredCubes.clear();
		for (Cube cube : cubes)
			uncoveredCubes.add(cube);
	}
	
	/**
	 * Rinitialiser le champ {@link #unusedParts}.
	 */
	public void resetUnusedParts() {
		unusedParts.clear();
		for (PuzzlePart puzzlePart : PuzzlePart.values())
			unusedParts.add(puzzlePart);
	}
	
	/**
	 * Construit la reprsentation d'un puzzle  partir d'une source d'informations
	 * elle-mme contenue dans un objet <i>Scanner</i> des API standards Java.
	 * <p>
	 * Cette approche prsente l'avantage de ne pas avoir  tenir compte du type de
	 * provenance des informations  analyser (fichier, chane de caractre, flux, ...).
	 * 
	 * @param source un objet <i>Scanner</i> qui contient les informations 
	 * analyser
	 * 
	 * @throws IllegalArgumentException si la source fournie ne contient pas une 
	 * dfinition de puzzle valide
	 */
	public void loadFromScanner(Scanner source) {
		int i = 0;
		// Va tre utilis pour reprer les doublons "en O(1)" :
		Set<Cube> cubesSet = new LinkedHashSet<Cube>(CUBES_NUMBER);

		while (i < CUBES_NUMBER) {
			if ( ! source.hasNextLine() )
				throw new IllegalArgumentException("Il manque "+(CUBES_NUMBER-i)+" lignes dans la source");
			computeLine(source.nextLine(), cubesSet);
			i++;
		}
		
		if (source.hasNextLine())
			throw new IllegalArgumentException("La source contient plus de 27 lignes");
	
		// Maintenant que tout est contrl, on peut charger les points dans une 
		// structure de donnes plus performante :
		i = 0;
		for (Cube p : cubesSet) {
			cubes[i] = p;
			i++;
		}
	}

	/**
	 * Traite les informations contenues sur une ligne de la source de
	 * dfintion de puzzle.
	 * 
	 * @param lineScanner contient la ligne  analyser.
	 * @param pointsSet structure permettant de contrler facilement l'unicit de
	 * coordonnes de chaque cube du puzzle
	 * 
	 * @throws IllegalArgumentException si la ligne  analyser n'est pas
	 * conforme au format demand ou si un doublon est repr
	 */
	private void computeLine(String line, Set<Cube> pointsSet) {
		if ( ! line.matches(SOURCE_LINE_PATTERN) )
			throw new IllegalArgumentException("La ligne suivante n'a pas une syntaxe valide : "+line);
		
		Scanner lineScanner = new Scanner(line);
		lineScanner.useDelimiter(",");
		
		Cube p = new Cube();
		p.x = lineScanner.nextInt();
		p.y = lineScanner.nextInt();
		p.z = lineScanner.nextInt();
		if (pointsSet.contains(p))
			throw new IllegalArgumentException("Le point de coordonnes ("+p.x+", "+p.y+", "+p.z+") figure deux fois dans le puzzle");
		
		pointsSet.add(p);
	}
}
